#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <time.h>

#include "md5.h"
#include "newinfi.h"

static size_t
datestring (char dates[], size_t dates_size, size_t *dates_len)
{
  time_t now;
  struct tm gmtime;
  size_t len;

  now = time (NULL);
  gmtime_r (&now, &gmtime);
  len = strftime (dates, dates_size, "%a, %d %b %Y %H:%M:%S GMT", &gmtime);
  if ( len == 0 )
    {
      dates[0] = 0;
      if ( dates_len )
	*dates_len = 0;
      return 0;
    }
  else
    {
      if ( dates_len )
	*dates_len = len;
      return len;
    }
}

static size_t
headers (char buf[], size_t buf_size, size_t *buf_len,
	 int code, const char codestr[], const char typestr[])
{
  char dates[256];

  datestring (dates, sizeof(dates), NULL);
  return rsprintf (buf, buf_size, buf_len,
		   "HTTP/1.1 %d %s\r\n"
		   "Date: %s\r\n"
		   "Content-Type: text/%s; charset=us-ascii\r\n"
		   "\r\n",
		   code, codestr,
		   dates, typestr);
}

static size_t
make_error (char buf[], size_t buf_size, size_t *buf_len,
	    int code, const char codestr[],
	    char body, const char explstr[])
{
  size_t len;

  headers (buf, buf_size, &len, code, codestr, "html");
  if ( body )
    len += rsprintf (buf+len, buf_size-len, NULL,
		     "<?xml version=\"1.0\" encoding=\"us-ascii\"?>\r\n\r\n"
		     "<!DOCTYPE html PUBLIC "
		     "\"-//W3C//DTD XHTML 1.0 Strict//EN\" "
		     "\"http://www.w3.org/TR/xhtml1/DTD/"
		     "xhtml1-strict.dtd\">\r\n\r\n"
		     "<html xmlns=\"http://www.w3.org/1999/xhtml\" "
		     "xml:lang=\"en\" lang=\"en\">\r\n"
		     "<head>\r\n"
		     "<title>%d %s</title>\r\n"
		     "</head>\r\n"
		     "<body>\r\n"
		     "<h1>%d %s</h1>\r\n"
		     "%s"
		     "</body>\r\n"
		     "</html>\r\n",
		     code, codestr, code, codestr, explstr);
  if ( buf_len )
    *buf_len = len;
  return len;
}

void
make_internal_error (char buf[], size_t buf_size, size_t *buf_len,
		     const char errstr[], char body)
{
  char explstr[256];

  rsprintf (explstr, sizeof(explstr), NULL,
	    "<p>This server encountered an internal error while "
	    "processing your request.</p>\r\n"
	    "<p>%s</p>\r\n", errstr);
  make_error (buf, buf_size, buf_len, 500, "Internal Server Error",
	      body, explstr);
}

void
make_unimplemented (char buf[], size_t buf_size, size_t *buf_len, char body)
{
  make_error (buf, buf_size, buf_len, 501, "Unimplemented",
	      body, "<p>Your browser sent a request using a method "
	      "that this server does not implement.</p>\r\n");
}

void
make_bad_request (char buf[], size_t buf_size, size_t *buf_len, char body)
{
  make_error (buf, buf_size, buf_len, 400, "Bad Request",
	      body, "<p>Your browser sent a request that this server "
	      "could not understand.</p>\r\n");
}

void
make_bad_uri (char buf[], size_t buf_size, size_t *buf_len, char body)
{
  make_error (buf, buf_size, buf_len, 400, "Bad Request-URI",
	      body, "<p>Your browser sent a request that this server "
	      "could not understand (invalid URI).</p>\r\n");
}

void
make_request_timeout (char buf[], size_t buf_size, size_t *buf_len, char body)
{
  make_error (buf, buf_size, buf_len, 408, "Request Timeout",
	      body, "<p>Your browser did not send a request in the "
	      "allotted time.</p>\r\n");
}

void
make_request_too_long (char buf[], size_t buf_size, size_t *buf_len, char body)
{
  make_error (buf, buf_size, buf_len, 413, "Request Entity Too Large",
	      body, "<p>Your browser sent a request that is too long "
	      "for this server.</p>\r\n");
}

void
make_200 (char buf[], size_t buf_size, size_t *buf_len,
	  char body, const char content[], const char typestr[])
{
  size_t len;
  size_t i;
  struct md5_ctx ctx;
  unsigned char digest[18];
  static const char base64[64]
    = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
  static const char pad = '=';
  char digest64[25];

  headers (buf, buf_size, &len, 200, "OK", typestr);
  MD5_start (&ctx);
  for ( i=0 ; content[i] ; i++ )
    MD5_feed (&ctx, content[i]);
  MD5_stop (&ctx, digest);
  digest[16] = 0;
  digest[17] = 0;
  for ( i=0 ; i<6 ; i++ )
    {
      digest64[i*4] = base64[(digest[i*3]>>2)&0x3f];
      digest64[i*4+1] = base64[((digest[i*3]&0x3)<<4)
			       |((digest[i*3+1]>>4)&0xf)];
      digest64[i*4+2] = base64[((digest[i*3+1]&0xf)<<2)
			       |((digest[i*3+2]>>6)&0x3)];
      digest64[i*4+3] = base64[digest[i*3+2]&0x3f];
    }
  digest64[22] = pad;
  digest64[23] = pad;
  digest64[24] = 0;
  len -= 2;  /* Remove CRLF (yuck!). */
  len += rsprintf (buf+len, buf_size-len, NULL,
		   "Content-Length: %ld\r\n"
		   "Content-MD5: %s\r\n\r\n",
		   (long)strlen(content),
		   digest64);
  if ( body )
    len += rsprintf (buf+len, buf_size-len, NULL,
		     "%s", content);
  if ( buf_len )
    *buf_len = len;
}

void
make_robots_txt (char buf[], size_t buf_size, size_t *buf_len,
		 char body)
{
  make_200 (buf, buf_size, buf_len, body,
	    "User-agent: *\r\nDisallow: /\r\n", "plain");
}
